from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.datastructures import Mesh


__all__ = ['Diagram']


class Diagram(Mesh):
    """Base diagram implementing attributes shared between the form and force diagram."""


# ==============================================================================
# Main
# ==============================================================================

if __name__ == "__main__":
    pass
